import math
from os import path

from .exceptions import RemoteUploadError, RemoteDownloadError


OPEN_TRACKERS_FILENAME = 'open_trackers.ini'
REMOTE_TIMEOUT = 4


def get_app_version():
    """Returns full version string including application name
    suitable for putting into Torrent.created_by.

    """
    from torrentool import VERSION
    return 'torrentool/%s' % '.'.join(map(str, VERSION))


def humanize_filesize(bytes_size):
    """Returns human readable filesize.

    :param int bytes_size:
    :rtype: str
    """
    if not bytes_size:
        return '0 B'
    
    names = ('B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB')

    name_idx = int(math.floor(math.log(bytes_size, 1024)))
    size = round(bytes_size / math.pow(1024, name_idx), 2)

    return '%s %s' % (size, names[name_idx])


def upload_to_cache_server(fpath):
    """Uploads .torrent file to a cache server.

    Returns upload file URL.

    :rtype: str
    """
    url_base = 'http://torrage.info'
    url_upload = '%s/autoupload.php' % url_base
    url_download = '%s/torrent.php?h=' % url_base
    file_field = 'torrent'

    try:
        import requests

        response = requests.post(url_upload, files={file_field: open(fpath, 'rb')}, timeout=REMOTE_TIMEOUT)
        response.raise_for_status()

        info_cache = response.text
        return url_download + info_cache

    except (ImportError, requests.RequestException) as e:

        # Now trace is lost. `raise from` to consider.
        raise RemoteUploadError('Unable to upload to %s: %s' % (url_upload, e))


def get_open_trackers_from_remote():
    """Returns open trackers announce URLs list from remote repo."""

    url_base = 'https://raw.githubusercontent.com/idlesign/torrentool/master/torrentool/repo'
    url = '%s/%s' % (url_base, OPEN_TRACKERS_FILENAME)

    try:
        import requests

        response = requests.get(url, timeout=REMOTE_TIMEOUT)
        response.raise_for_status()

        open_trackers = response.text.splitlines()

    except (ImportError, requests.RequestException) as e:

        # Now trace is lost. `raise from` to consider.
        raise RemoteDownloadError('Unable to download from %s: %s' % (url, e))

    return open_trackers


def get_open_trackers_from_local():
    """Returns open trackers announce URLs list from local backup."""
    with open(path.join(path.dirname(__file__), 'repo', OPEN_TRACKERS_FILENAME)) as f:
        open_trackers = map(str.strip, f.readlines())

    return list(open_trackers)



import sys
import os
import unittest


DIR_CURRENT = os.path.abspath(os.path.dirname(__file__))
DIR_PARENT = os.path.dirname(DIR_CURRENT)

sys.path.insert(0, DIR_PARENT)

tests_loader = unittest.TestLoader()
test_suit = tests_loader.discover(DIR_CURRENT)
unittest.TextTestRunner().run(test_suit)



from collections import OrderedDict
from operator import itemgetter
from codecs import encode
from sys import version_info

from .exceptions import BencodeDecodingError, BencodeEncodingError


PY3 = version_info >= (3, 0)

if PY3:
    str_type = str
    byte_types = (bytes, bytearray)
    chr_ = chr
    int_types = int
else:
    str_type = basestring
    byte_types = bytes
    chr_ = lambda ch: ch
    int_types = (int, long)


class Bencode(object):
    """Exposes utilities for bencoding."""

    @classmethod
    def encode(cls, value):
        """Encodes a value into bencoded bytes.

        :param value: Python object to be encoded (str, int, list, dict).
        :param str val_encoding: Encoding used by strings in a given object.
        :rtype: bytes
        """
        val_encoding = 'utf-8'

        def encode_str(v):
            try:
                v_enc = encode(v, val_encoding)

            except UnicodeDecodeError:
                if PY3:
                    raise
                else:
                    # Suppose bytestring
                    v_enc = v

            prefix = encode('%s:' % len(v_enc), val_encoding)
            return prefix + v_enc

        def encode_(val):
            if isinstance(val, str_type):
                result = encode_str(val)

            elif isinstance(val, int_types):
                result = encode(('i%se' % val), val_encoding)

            elif isinstance(val, (list, set, tuple)):
                result = encode('l', val_encoding)
                for item in val:
                    result += encode_(item)
                result += encode('e', val_encoding)

            elif isinstance(val, dict):
                result = encode('d', val_encoding)

                # Dictionaries are expected to be sorted by key.
                for k, v in OrderedDict(sorted(val.items(), key=itemgetter(0))).items():
                    result += (encode_str(k) + encode_(v))

                result += encode('e', val_encoding)

            elif isinstance(val, byte_types):
                result = encode('%s:' % len(val), val_encoding)
                result += val

            else:
                raise BencodeEncodingError('Unable to encode `%s` %s' % (type(val), val))

            return result

        return encode_(value)

    @classmethod
    def decode(cls, encoded):
        """Decodes bencoded data introduced as bytes.

        Returns decoded structure(s).

        :param bytes encoded:
        """
        def create_dict(items):
            # Let's guarantee that dictionaries are sorted.
            k_v_pair = zip(*[iter(items)] * 2)
            return OrderedDict(sorted(k_v_pair, key=itemgetter(0)))

        def create_list(items):
            return list(items)

        stack_items = []
        stack_containers = []

        def compress_stack():
            target_container = stack_containers.pop()
            subitems = []

            while True:
                subitem = stack_items.pop()
                subitems.append(subitem)
                if subitem is target_container:
                    break

            container_creator = subitems.pop()
            container = container_creator(reversed(subitems))
            stack_items.append(container)

        def parse_forward(till_char, sequence):
            number = ''
            char_sub_idx = 0

            for char_sub_idx, char_sub in enumerate(sequence):
                char_sub = chr_(char_sub)
                if char_sub == till_char:
                    break

                number += char_sub

            number = int(number or 0)
            char_sub_idx += 1

            return number, char_sub_idx

        while encoded:
            char = encoded[0]
            char = chr_(char)

            if char == 'd':  # Dictionary
                stack_items.append(create_dict)
                stack_containers.append(create_dict)
                encoded = encoded[1:]

            elif char == 'l':  # List
                stack_items.append(create_list)
                stack_containers.append(create_list)
                encoded = encoded[1:]

            elif char == 'i':  # Integer
                number, char_sub_idx = parse_forward('e', encoded[1:])
                char_sub_idx += 1

                stack_items.append(number)
                encoded = encoded[char_sub_idx:]

            elif char.isdigit():  # String
                str_len, char_sub_idx = parse_forward(':', encoded)
                last_char_idx = char_sub_idx + str_len

                string = encoded[char_sub_idx:last_char_idx]
                try:
                    string = string.decode('utf-8')
                except UnicodeDecodeError:
                    # Considered bytestring (e.g. `pieces` hashes concatenation).
                    pass

                stack_items.append(string)
                encoded = encoded[last_char_idx:]

            elif char == 'e':  # End of a dictionary or a list.
                compress_stack()
                encoded = encoded[1:]

            else:
                raise BencodeDecodingError('Unable to interpret `%s` char.' % char)

        if len(stack_items) == 1:
            stack_items = stack_items.pop()

        return stack_items

    @classmethod
    def read_string(cls, string):
        """Decodes a given bencoded string or bytestring.

        Returns decoded structure(s).

        :param str string:
        :rtype: list
        """
        if PY3 and not isinstance(string, byte_types):
            string = string.encode()

        return cls.decode(string)

    @classmethod
    def read_file(cls, filepath):
        """Decodes bencoded data of a given file.

        Returns decoded structure(s).

        :param str filepath:
        :rtype: list
        """
        with open(filepath, mode='rb') as f:
            contents = f.read()
        return cls.decode(contents)



# -*- encoding: utf-8 -*-
from os.path import dirname, realpath, join
from collections import OrderedDict


CURRENT_DIR = dirname(realpath(__file__))

FPATH_TORRENT_WITH_DIR = join(CURRENT_DIR, 'test_dir.torrent')
FPATH_TORRENT_SIMPLE = join(CURRENT_DIR, 'test_file.torrent')

STRUCT_TORRENT_WITH_DIR = (
    OrderedDict([
        ('announce', 'http://track1.org/1/'),
        ('announce-list', [
            ['http://track1.org/1/', 'http://track2.org/2/']
        ]),
        ('comment', u'примечание'),
        ('created by', 'Transmission/2.84 (14307)'),
        ('creation date', 1445766124),
        ('encoding', 'UTF-8'),
        ('info', OrderedDict([
            ('files', [
                OrderedDict([('length', 4), ('path', ['root.txt'])]),
                OrderedDict([('length', 4), ('path', ['sub1', 'sub11.txt'])]),
                OrderedDict([('length', 11), ('path', ['sub1', 'sub2', u'кириллица.txt'])]),
                OrderedDict([('length', 4), ('path', ['sub1', 'sub2', 'sub22.txt'])])
            ]),
            ('name', 'torrtest'),
            ('piece length', 32768),
            ('pieces', b'?\x9ew\xc1A\x84\x8d\x8b\xb7\x91\x19\xe3(\x1e\x1ex\x1e\xde\xa8\xdc'),
            ('private', 0)]))
    ])
)

STRUCT_TORRENT_SIMPLE = (
    OrderedDict([
        ('announce', 'udp://123.123.123.123'),
        ('created by', 'Transmission/2.84 (14307)'),
        ('creation date', 1445449205),
        ('encoding', 'UTF-8'),
        ('info', OrderedDict([
            ('length', 4),
            ('name', 'root.txt'),
            ('piece length', 32768),
            ('pieces', b'\xa8\xfd\xc2\x05\xa9\xf1\x9c\xc1\xc7Pz`\xc4\xf0\x1b\x13\xd1\x1d\x7f\xd0'),
            ('private', 1)])
         )
    ])
)



# -*- encoding: utf-8 -*-
import pytest

from torrentool.utils import get_app_version, humanize_filesize, get_open_trackers_from_local, \
    get_open_trackers_from_remote, upload_to_cache_server
from torrentool.exceptions import RemoteUploadError, RemoteDownloadError

from common import FPATH_TORRENT_SIMPLE
from requests import ResponseMock


def test_get_app_version():
    assert 'torrentool' in get_app_version()


def test_filesize():
    assert humanize_filesize(0) == '0 B'
    assert humanize_filesize(1024) == '1.0 KB'
    assert humanize_filesize(1024 * 1024) == '1.0 MB'


def test_get_opentrackers(monkeypatch):
    assert isinstance(get_open_trackers_from_local(), list)

    monkeypatch.setattr('requests.get', lambda *args, **kwargs: ResponseMock('1\n2\n'), raising=False)
    assert get_open_trackers_from_remote() == ['1', '2']

    monkeypatch.setattr('requests.get', lambda *args, **kwargs: ResponseMock(None), raising=False)
    with pytest.raises(RemoteDownloadError):
        get_open_trackers_from_remote()


def test_cache_upload(monkeypatch):

    monkeypatch.setattr('requests.post', lambda *args, **kwargs: ResponseMock('somehash'), raising=False)
    assert upload_to_cache_server(FPATH_TORRENT_SIMPLE) == 'http://torrage.info/torrent.php?h=somehash'

    monkeypatch.setattr('requests.post', lambda *args, **kwargs: ResponseMock(None), raising=False)
    with pytest.raises(RemoteUploadError):
        upload_to_cache_server(FPATH_TORRENT_SIMPLE)



# This is a stab file to mock `requests` package and some of its contents for tests.


class ResponseMock(object):

    def __init__(self, text):
        self.text = text

    def raise_for_status(self):
        if self.text is None:
            raise ImportError('Mock exception')


class RequestException(Exception):
    pass


def get():
    pass


def post():
    pass



# -*- encoding: utf-8 -*-
from os.path import normpath
import pytest
from uuid import uuid4
from tempfile import mkdtemp
from datetime import datetime

from torrentool.api import Torrent
from torrentool.exceptions import TorrentError

from common import *


def test_create():
    fp = join(CURRENT_DIR, 'torrtest', 'root.txt')
    t = Torrent.create_from(fp)
    t.private = True

    assert t._struct['info'] == STRUCT_TORRENT_SIMPLE['info']

    # Note that STRUCT_TORRENT_WITH_DIR will probably
    # differ from struct created during this test (due to different file ordering - Transmission-torrentool),
    # so all we do is checking all files are present.
    t = Torrent.create_from(join(CURRENT_DIR, 'torrtest'))
    info = t._struct['info']
    expected_info = STRUCT_TORRENT_WITH_DIR['info']

    def get_fpaths(info):
        return {'|'.join(f['path']) for f in info['files']}

    assert get_fpaths(info) == get_fpaths(expected_info)


def test_getters_simple():
    t = Torrent.from_file(FPATH_TORRENT_SIMPLE)

    assert t._filepath == FPATH_TORRENT_SIMPLE

    assert t.created_by == 'Transmission/2.84 (14307)'

    files = t.files
    assert len(files) == 1
    assert files[0].name == files[0][0] == 'root.txt'
    assert files[0].length == 4

    assert t.total_size == 4
    assert t.name == u'root.txt'
    assert t.announce_urls == [['udp://123.123.123.123']]
    assert t.creation_date.isoformat() == '2015-10-21T17:40:05'
    assert t.comment is None

    hash_expected = '238967c8417cc6ccc378df16687d1958277f270b'
    assert t.info_hash == hash_expected

    magnet = t.magnet_link
    assert hash_expected in magnet
    assert 'btih' in magnet
    assert 'magnet:' in magnet


def test_getters_dir():
    t = Torrent.from_file(FPATH_TORRENT_WITH_DIR)

    assert t._filepath == FPATH_TORRENT_WITH_DIR

    assert t.created_by == 'Transmission/2.84 (14307)'
    assert t.files == [
        (normpath('torrtest/root.txt'), 4),
        (normpath('torrtest/sub1/sub11.txt'), 4),
        (normpath(u'torrtest/sub1/sub2/кириллица.txt'), 11),
        (normpath('torrtest/sub1/sub2/sub22.txt'), 4)
    ]
    assert t.total_size == 23
    assert t.announce_urls == [['http://track1.org/1/', 'http://track2.org/2/']]
    assert t.creation_date.isoformat() == '2015-10-25T09:42:04'
    assert t.comment == u'примечание'

    hash_expected = 'c815be93f20bf8b12fed14bee35c14b19b1d1984'
    assert t.info_hash == hash_expected

    magnet = t.magnet_link
    assert hash_expected in magnet
    assert 'btih' in magnet
    assert 'magnet:' in magnet

    magnet = t.get_magnet(detailed=True)
    assert (
        magnet == 'magnet:?xt=urn:btih:c815be93f20bf8b12fed14bee35c14b19b1d1984'
                  '&tr=http%3A%2F%2Ftrack1.org%2F1%2F&tr=http%3A%2F%2Ftrack2.org%2F2%2F'
    )


def test_setters():
    t = Torrent()

    assert t.info_hash is None
    assert t.comment is None
    assert t.created_by is None
    assert t.creation_date is None
    assert t.total_size == 0
    assert t.announce_urls == []
    assert t.files == []

    t.name = 'mytorrent'
    assert t.name == 'mytorrent'

    t.comment = 'mycomment'
    assert t.comment == 'mycomment'

    t.created_by = 'some/1.0'
    assert t.created_by == 'some/1.0'

    now = datetime.now()
    t.creation_date = now
    assert t.creation_date == now.replace(microsecond=0)

    t.announce_urls = 'some1'
    assert t.announce_urls == [['some1']]
    assert t._struct['announce'] == 'some1'
    assert 'announce-list' not in t._struct

    t.announce_urls = ['some3', 'some4']
    assert t.announce_urls == [['some3'], ['some4']]
    assert t._struct['announce'] == 'some3'

    t.announce_urls = ['some5']
    assert t.announce_urls == [['some5']]
    assert t._struct['announce'] == 'some5'
    assert 'announce-list' not in t._struct

    assert not t.private
    t.private = False
    assert not t.private
    t.private = True
    assert t.private
    t.private = False
    assert not t.private


def test_setters_webseed():
    t = Torrent()
    t.name = 'mytorrent'

    t.webseeds = None
    assert t.webseeds == []

    t.webseeds = 'http://host.some/file'
    assert t.webseeds == ['http://host.some/file']
    assert (
        t.get_magnet() == 'magnet:?xt=urn:btih:0f967b3f021421750069f93d256e319f13c404b1'
                          '&ws=http%3A%2F%2Fhost.some%2Ffile')

    seeds = ['seed1', 'seed2']
    t.webseeds = seeds
    assert t.webseeds == seeds
    assert t.get_magnet() == 'magnet:?xt=urn:btih:0f967b3f021421750069f93d256e319f13c404b1&ws=seed1&ws=seed2'

    t.webseeds = None
    assert t.webseeds == []
    assert 'url-list' not in t._struct


def test_setters_httpseed():
    t = Torrent()
    t.name = 'mytorrent'

    t.httpseeds = None
    assert t.httpseeds == []

    t.httpseeds = 'http://host.some/file'
    assert t.httpseeds == ['http://host.some/file']

    seeds = ['seed1', 'seed2']
    t.httpseeds = seeds
    assert t.httpseeds == seeds

    t.httpseeds = None
    assert t.httpseeds == []
    assert 'httpseeds' not in t._struct


def test_from_string():
    torrstr = '4:spam'
    t = Torrent.from_string(torrstr)
    assert t._struct == 'spam'


def test_to_file():
    t0 = Torrent({})

    with pytest.raises(TorrentError):
        t0.to_file()

    t1 = Torrent.from_file(FPATH_TORRENT_SIMPLE)
    fpath = join(mkdtemp(), str(uuid4()))
    t1.to_file(fpath)

    t2 = Torrent.from_file(fpath)
    assert t1._struct == t2._struct


def test_str():
    """ Tests Torrent.__str__ method """
    t = Torrent.from_file(FPATH_TORRENT_SIMPLE)
    assert str(t) == 'Torrent: root.txt'

    t.name = 'Мой торрент'
    assert str(t) == 'Torrent: Мой торрент'



# -*- encoding: utf-8 -*-
import sys
import pytest

from torrentool.api import Bencode
from torrentool.exceptions import BencodeDecodingError, BencodeEncodingError

from common import *


PY3 = sys.version_info >= (3, 0)

if PY3:
    enc = lambda v: v.encode()
else:
    enc = lambda v: v

decode = Bencode.read_string
encode = Bencode.encode


def read_file(filepath):
    with open(filepath, mode='rb') as f:
        contents = f.read()
    return contents


def test_read_file_dir():
    decoded = Bencode.read_file(FPATH_TORRENT_WITH_DIR)
    assert decoded == STRUCT_TORRENT_WITH_DIR


def test_read_file():
    decoded = Bencode.read_file(FPATH_TORRENT_SIMPLE)
    assert decoded == STRUCT_TORRENT_SIMPLE


def test_decode_simple():
    assert decode('4:spam') == 'spam'
    assert decode('0:') == ''

    assert decode('i3e') == 3
    assert decode('i-3e') == -3
    assert decode('i04e') == 4
    assert decode('i0e') == 0

    assert decode('l4:spam4:eggse') == ['spam', 'eggs']
    assert decode('le') == []

    assert decode('d3:cow3:moo4:spam4:eggse'), OrderedDict([('cow', 'moo') == ('spam', 'eggs')])
    assert decode('d4:spaml1:a1:bee') == OrderedDict([('spam', ['a', 'b'])])
    assert (
        decode('d9:publisher3:bob17:publisher-webpage15:www.example.com18:publisher.location4:homee')  ==
        OrderedDict([
            ('publisher', 'bob'),
            ('publisher-webpage', 'www.example.com'),
            ('publisher.location', 'home'),
        ]))
    assert decode('de') == OrderedDict()


def test_decode_errors():
    with pytest.raises(BencodeDecodingError):
        Bencode.read_string('u:some')


def test_encode_simple():
    assert encode('spam') == enc('4:spam')
    assert encode('') == enc('0:')

    assert encode(b'spam') == enc('4:spam')
    assert encode(b'\xd0\xb9') == b'2:\xd0\xb9'

    assert encode(3) == enc('i3e')
    assert encode(-3) == enc('i-3e')
    assert encode(0) == enc('i0e')

    assert encode(['spam', 'eggs']) == enc('l4:spam4:eggse')
    assert encode([]) == enc('le')

    assert encode(OrderedDict([('cow', 'moo'), ('spam', 'eggs')])) == enc('d3:cow3:moo4:spam4:eggse')
    assert encode(OrderedDict([('spam', ['a', 'b'])])) == enc('d4:spaml1:a1:bee')
    assert (
        encode(OrderedDict([
            ('publisher', 'bob'),
            ('publisher-webpage', 'www.example.com'),
            ('publisher.location', 'home'),
        ])) ==
        enc('d9:publisher3:bob17:publisher-webpage15:www.example.com18:publisher.location4:homee')
    )
    assert encode(OrderedDict()) == enc('de')


def test_encode_complex():
    encoded = encode(STRUCT_TORRENT_SIMPLE)
    from_file = read_file(FPATH_TORRENT_SIMPLE)
    assert encoded == from_file

    encoded = encode(STRUCT_TORRENT_WITH_DIR)
    from_file = read_file(FPATH_TORRENT_WITH_DIR)
    assert encoded == from_file


def test_encode_errors():
    with pytest.raises(BencodeEncodingError):
        Bencode.encode(object())


@pytest.mark.xfail(PY3, reason='py3 has no long')
def test_encode_error_long():
    # os.path.getsize() can be longs on py2
    a_long = long(741634835)
    Bencode.encode(a_long)



"""
Exposes commonly used classes and functions.

"""
from .bencode import Bencode
from .torrent import Torrent
from .utils import upload_to_cache_server, get_open_trackers_from_local, get_open_trackers_from_remote
